@@ -1,15 +1,7 @@
 /*
- * Copyright (C) by Roeland Jago Douma <roeland@famdouma.nl>
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
- * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
- * for more details.
+ * SPDX-FileCopyrightText: 2017 Nextcloud GmbH and Nextcloud contributors
+ * SPDX-FileCopyrightText: 2015 ownCloud GmbH
+ * SPDX-License-Identifier: GPL-2.0-or-later
  */
 
 #include "capabilities.h"
@@ -210,8 +202,9 @@ bool Capabilities::isClientStatusReportingEnabled() const
 
 QList<QByteArray> Capabilities::supportedChecksumTypes() const
 {
-    QList<QByteArray> list;
-    foreach (const auto &t, _capabilities["checksums"].toMap()["supportedTypes"].toList()) {
+    const auto supportedTypes = _capabilities["checksums"].toMap()["supportedTypes"].toList();
+    QList<QByteArray> list(supportedTypes.count());
+    for (const auto &t : supportedTypes) {
         list.push_back(t.toByteArray());
     }
     return list;
@@ -245,6 +238,16 @@ bool Capabilities::chunkingNg() const
     return _capabilities["dav"].toMap()["chunking"].toByteArray() >= "1.0";
 }
 
+qint64 Capabilities::maxChunkSize() const
+{
+    return _capabilities["files"].toMap()["chunked_upload"].toMap()["max_size"].toLongLong();
+}
+
+int Capabilities::maxConcurrentChunkUploads() const
+{
+    return _capabilities["files"].toMap()["chunked_upload"].toMap()["max_parallel_count"].toInt();
+}
+
 bool Capabilities::bulkUpload() const
 {
     return _capabilities["dav"].toMap()["bulkupload"].toByteArray() >= "1.0";
@@ -362,8 +365,9 @@ bool Capabilities::privateLinkPropertyAvailable() const
 
 QList<int> Capabilities::httpErrorCodesThatResetFailingChunkedUploads() const
 {
-    QList<int> list;
-    foreach (const auto &t, _capabilities["dav"].toMap()["httpErrorCodesThatResetFailingChunkedUploads"].toList()) {
+    const auto httpErrorCodes = _capabilities["dav"].toMap()["httpErrorCodesThatResetFailingChunkedUploads"].toList();
+    QList<int> list(httpErrorCodes.count());
+    for (const auto &t : httpErrorCodes) {
         list.push_back(t.toInt());
     }
     return list;
@@ -435,7 +439,7 @@ void Capabilities::addDirectEditor(DirectEditor* directEditor)
 
 DirectEditor* Capabilities::getDirectEditorForMimetype(const QMimeType &mimeType)
 {
-    foreach(DirectEditor* editor, _directEditors) {
+    for (const auto editor : std::as_const(_directEditors)) {
         if(editor->hasMimetype(mimeType))
             return editor;
     }
@@ -445,7 +449,7 @@ DirectEditor* Capabilities::getDirectEditorForMimetype(const QMimeType &mimeType
 
 DirectEditor* Capabilities::getDirectEditorForOptionalMimetype(const QMimeType &mimeType)
 {
-    foreach(DirectEditor* editor, _directEditors) {
+    for (const auto editor : std::as_const(_directEditors)) {
         if(editor->hasOptionalMimetype(mimeType))
             return editor;
     }
